/*
**                    TMM v1.0 for Avisynth 2.5.x
**
**   TMM builds a motion-mask for TDeint, which TDeint uses via
**   its 'emask' parameter.  TMM can use fixed or per-pixel adaptive
**   motion thresholds, as well as any length static period.  It
**   checks backwards, across, and forwards when looking for motion.
**
**   Copyright (C) 2007 Kevin Stone
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 2 of the License, or
**   (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "TMM.h"

AVSValue __cdecl Create_TMM(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	if (!args[0].IsClip())
		env->ThrowError("TMM:  first argument must be a clip!");
	PClip oclip = args[0].AsClip();
	if (oclip->GetVideoInfo().IsFieldBased())
		env->ThrowError("TMM:  input clip cannot be field based!");
	int mode = args[1].AsInt(0);
	if (mode != 0 && mode != 1)
		env->ThrowError("TMM:  mode must be set to 0 or 1!");
	int order = args[2].AsInt(-1);
	if (order < -1 || order > 1)
		env->ThrowError("TMM:  order must be set to -1, 0, or 1!");
	if (order == -1)
		order = oclip->GetParity(0) ? 1 : 0;
	int field = args[3].AsInt(-1);
	if (field < -1 || field > 1)
		env->ThrowError("TMM:  field must be set to -1, 0, or 1!");
	if (field == -1)
		field = order;
	try
	{
		oclip = env->Invoke("SeparateFields",oclip).AsClip();
		PClip topf, botf;
		if (oclip->GetParity(0))
		{
			topf = env->Invoke("SelectEven",oclip).AsClip();
			botf = env->Invoke("SelectOdd",oclip).AsClip();
		}
		else
		{
			topf = env->Invoke("SelectOdd",oclip).AsClip();
			botf = env->Invoke("SelectEven",oclip).AsClip();
		}
		topf = new createMM(topf,args[6].AsInt(1),args[7].AsInt(-1),
			args[8].AsInt(-1),args[9].AsInt(-1),args[10].AsInt(-1),
			args[11].AsInt(2),args[12].AsInt(4),args[13].AsInt(75),
			args[14].AsInt(4),env);
		botf = new createMM(botf,args[6].AsInt(1),args[7].AsInt(-1),
			args[8].AsInt(-1),args[9].AsInt(-1),args[10].AsInt(-1),
			args[11].AsInt(2),args[12].AsInt(4),args[13].AsInt(75),
			args[14].AsInt(4),env);
		return new buildMM(topf,botf,mode,order,field,args[4].AsInt(10),
			args[5].AsInt(1),env);
	}
	catch (IScriptEnvironment::NotFound)
	{
		env->ThrowError("TMM:  IScriptEnvironment::NotFound error while" \
			" invoking filters!");
	}
	catch (AvisynthError e)
	{
		env->ThrowError("TMM:  avisynth error while invoking filters (%s)!", 
			e.msg);
	}
	return 0;
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env) 
{
	env->AddFunction("TMM","c[mode]i[order]i[field]i[length]i[mtype]i[ttype]i[mtqL]i" \
		"[mthL]i[mtqC]i[mthC]i[nt]i[minthresh]i[maxthresh]i[cstr]i", Create_TMM, 0);
    return 0;
}