/*
**                    TMM v1.0 for Avisynth 2.5.x
**
**   TMM builds a motion-mask for TDeint, which TDeint uses via
**   its 'emask' parameter.  TMM can use fixed or per-pixel adaptive
**   motion thresholds, as well as any length static period.  It
**   checks backwards, across, and forwards when looking for motion.
**
**   Copyright (C) 2007 Kevin Stone
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 2 of the License, or
**   (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <windows.h>
#include "avisynth.h"
#include "PlanarFrame.h"

const int vlut2a[64] = {
	0, 1, 2, 2, 3, 0, 2, 2,
	1, 1, 2, 2, 0, 1, 2, 2,
	2, 2, 2, 2, 2, 2, 2, 2,
	2, 2, 2, 2, 2, 2, 2, 2,
	3, 0, 2, 2, 3, 3, 2, 2,
	0, 1, 2, 2, 3, 1, 2, 2,
	2, 2, 2, 2, 2, 2, 2, 2,
	2, 2, 2, 2, 2, 2, 2, 2 };

const int vlut2b[64] = {
	0, 0, 2, 2, 0, 0, 2, 2,
	0, 1, 2, 2, 0, 1, 2, 2,
	2, 2, 2, 2, 2, 2, 2, 2,
	2, 2, 2, 2, 2, 2, 2, 2,
	0, 0, 2, 2, 3, 3, 2, 2,
	0, 1, 2, 2, 3, 1, 2, 2,
	2, 2, 2, 2, 2, 2, 2, 2,
	2, 2, 2, 2, 2, 2, 2, 2 };

const int vlut3[64] = {
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 1, 0, 1, 0, 1, 0, 1,
	0, 0, 2, 2, 0, 0, 2, 2,
	0, 1, 2, 2, 0, 1, 2, 2,
	0, 0, 0, 0, 3, 3, 3, 3,
	0, 1, 0, 1, 3, 1, 3, 1,
	0, 0, 2, 2, 3, 3, 2, 2,
	0, 1, 2, 2, 3, 1, 2, 2 };

const int tmmlut16[2][8] = {
	60, 20, 50, 10, 60, 10, 40, 30,
	60, 10, 40, 30, 60, 20, 50, 10 };

struct PFCB {
	PlanarFrame *pf;
	int n;
};

class buildMM : public GenericVideoFilter
{
private:
	PClip bfields;
	VideoInfo vi_saved;
	int order, field, mode, length, mtype;
	int *gvlut;
	const int *vlut;
	PFCB **srcPFb, **srcPFt;
	PlanarFrame *dstPF;
	void buildMM::setBlack(PlanarFrame *p);
	void buildMM::resetFrames(PFCB **p, int start, int stop);

public:
	PVideoFrame __stdcall buildMM::GetFrame(int n, IScriptEnvironment *env);
	buildMM::buildMM(PClip _tfields, PClip _bfields, int _mode, int _order, 
		int _field, int _length, int _mtype, IScriptEnvironment *env);
	buildMM::~buildMM();
};