//    a60224 filter for Avisynth.
//    Copyright (C) 2003-2004 chiyo-clone
//    http://www.chiyoclone.net/
//    master@chiyoclone.net

//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.

//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <math.h>
#include <float.h>
#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include "avisynth.h"

//#define LARGE_VAL (1.7E+308)

#define BITS(b7,b6,b5,b4,b3,b2,b1,b0) (((b7)<<7)|((b6)<<6)|((b5)<<5)|((b4)<<4)|((b3)<<3)|((b2)<<2)|((b1)<<1)|(b0))

#ifdef _TEST
#include "font.cpp"
#endif

#include "table.cpp"

class a60224_common : public GenericVideoFilter{
protected:
	double blur_rate;
	int crop_x1, crop_x2, crop_y1, crop_y2;
	
	double *p_d, *p_d2;
	unsigned char *p_b;
	unsigned char *p_m;
	long *p_ns;
#ifdef _TEST
	unsigned long p_debug_fnc[64];
	unsigned char *p_debug_fnb;
#endif
	PVideoFrame vfcache[12];
	int vfcache_i;
	void set_cache(PVideoFrame c){
		vfcache[vfcache_i] = c;
		vfcache_i = (vfcache_i+1)%12;
	}
public:
	a60224_common(
		PClip _child, double _blur_rate,
		int _crop_x1, int _crop_x2, int _crop_y1, int _crop_y2
	) : GenericVideoFilter(_child), vfcache_i(0){
		int i;
		p_d = (double*)malloc(sizeof(double)*(vi.num_frames+8));
		for(i=0; i<vi.num_frames+4; i++) p_d[i] = HUGE_VAL;
		p_d += 4;
		p_d[-4] = 0;
		p_d[-3] = 0;
		p_d[-2] = 0;
		p_d[-1] = 0;
		p_d[vi.num_frames] = 0;
		p_d[vi.num_frames+1] = 0;
		p_d[vi.num_frames+2] = 0;
		p_d[vi.num_frames+3] = 0;
		p_d2 = (double*)malloc(sizeof(double)*(vi.num_frames+8));
		for(i=0; i<vi.num_frames+8; i++) p_d2[i] = HUGE_VAL;
		p_d2 += 4;
		p_d2[-4] = 0;
		p_d2[-3] = 0;
		p_d2[-2] = 0;
		p_d2[-1] = 0;
		p_d2[vi.num_frames] = 0;
		p_d2[vi.num_frames+1] = 0;
		p_d2[vi.num_frames+2] = 0;
		p_d2[vi.num_frames+3] = 0;
		blur_rate = _blur_rate;
		crop_x1 = _crop_x1;
		crop_x2 = _crop_x2;
		crop_y1 = _crop_y1;
		crop_y2 = _crop_y2;
		p_b = (unsigned char*)malloc(vi.num_frames+4);
		for(i=0; i<vi.num_frames+4; i++) p_b[i] = (unsigned char)0xFF;
		p_b += 2;
		p_b[-2] = 0;
		p_b[-1] = 0;
		p_b[vi.num_frames] = 0;
		p_b[vi.num_frames+1] = 0;
#ifdef _TEST
		p_debug_fnb = (unsigned char*)malloc(vi.num_frames);
		memset(p_debug_fnb, 0, vi.num_frames);
		for(i=0; i<64; i++) p_debug_fnc[i]=0;
#endif
		
		vi.num_frames = vi.num_frames * 2 / 5;
		vi.SetFPS(vi.fps_numerator*2, vi.fps_denominator*5);
		
		p_m = (unsigned char*)malloc(vi.num_frames+2);
		for(i=0; i<vi.num_frames+2; i++) p_m[i] = (unsigned char)0xFF;
		p_m += 1;
		p_m[-1] = 0;
		p_m[vi.num_frames] = 0;
		p_ns = (long*)malloc(sizeof(long)*(vi.num_frames));
		for(i=0; i<vi.num_frames; i++) p_ns[i] = 0x7FFFFFFF;
	}
	~a60224_common(){
		free(p_d-4);
		free(p_d2-4);
		free(p_b-2);
#ifdef _TEST
		free(p_debug_fnb);
#endif
		free(p_m-1);
		free(p_ns);
	}
	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env){
		if(n >= vi.num_frames){
			n = vi.num_frames - 1;
		}else if(n < 0){
			n = 0;
		}
		
		calc_ns(n, env);
		PVideoFrame c = child->GetFrame(p_ns[n], env);
		_asm{ emms }
#ifdef _TEST
		if(!(vi.IsYUY2()||vi.IsYV12()||vi.IsRGB24()||vi.IsRGB32())){
			env->ThrowError("a60224(TEST): Unsupported color format.");
		}
		
		const int n0 = n*5/2;
		
		unsigned char bb2, bb1, b0, bf1, bf2;
		calc_b(n0-2, env);
		calc_b(n0-1, env);
		calc_b(n0, env);
		calc_b(n0+1, env);
		calc_b(n0+2, env);
		bb2 = p_b[n0-2];
		bb1 = p_b[n0-1];
		b0 = p_b[n0];
		bf1 = p_b[n0+1];
		bf2 = p_b[n0+2];
		
		unsigned int i=0;
		i |= bb2?(1<<5):0;
		i |= bb1?(1<<4):0;
		i |= b0 ?(1<<3):0;
		i |= bf1?(1<<2):0;
		i |= bf2?(1<<1):0;
		i |= (n%2)?1:0;
		
		unsigned char mb1, m0, mf1, m;
		mb1 = p_m[n-1];
		m0 = p_m[n];
		mf1 = p_m[n+1];
		if(n%2){
			mb1 = ~(mb1<<2);
			mf1 = ~(mf1>>3);
		}else{
			mb1 = ~(mb1<<3);
			mf1 = ~(mf1>>2);
		}
		m = mb1 & m0 & mf1;
		if(!m) m = mb1 & m0;
		if(!m) m = m0 & mf1;
		if(!m) m = m0;
		
		calc_ns(n, env);
		int ns = p_ns[n];
		if(!p_debug_fnb[n]){
			p_debug_fnb[n] = true;
			p_debug_fnc[i]++;
		}
		if(!(env->MakeWritable(&c))){
			env->ThrowError("NO1");
		}
		if(!c->IsWritable()){
			env->ThrowError("NO2");
		}
		char buf[4096]; int buf_n = 0;
		sprintf(buf, "i=%2.1u m0=%2.2XH m=%2.2XH n0=%6.6u ns=%6.6u n=%6.6u", i, m0, m, n0, ns, n); write_str(c, buf, buf_n++);
		sprintf(buf, "%u%u%u%u%u", bb2?1:0, bb1?1:0, b0?1:0, bf1?1:0, bf2?1:0); write_str(c, buf, buf_n++);
		sprintf(buf, "-2 %f", p_d[n0-2]); write_str(c, buf, buf_n++);
		sprintf(buf, "-1 %f", p_d[n0-1]); write_str(c, buf, buf_n++);
		sprintf(buf, " 0 %f", p_d[n0]); write_str(c, buf, buf_n++);
		sprintf(buf, "+1 %f", p_d[n0+1]); write_str(c, buf, buf_n++);
		sprintf(buf, "+2 %f", p_d[n0+2]); write_str(c, buf, buf_n++);
		if(n == 0){
			sprintf(buf, " 0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ.-=+"); write_str(c, buf, buf_n++);
		}
		if(n == vi.num_frames-1){
			int i;
			for(i=0; i<16; i++){
				sprintf(buf, "%2.2u %4.1u  %2.2u %4.1u  %2.2u %4.1u  %2.2u %4.1u",
					i+0 , p_debug_fnc[i+0],
					i+16, p_debug_fnc[i+16],
					i+32, p_debug_fnc[i+32],
					i+48, p_debug_fnc[i+48]
				);
				write_str(c, buf, buf_n++);
			}
		}
#endif
		return c;
	}
protected:
	void calc_ns(int n, IScriptEnvironment* env){
		if(p_ns[n] != 0x7FFFFFFF) return;
		
		const int n0 = n*5/2;
		calc_m(n-1, env);
		calc_m(n, env);
		calc_m(n+1, env);
		unsigned char mb1, m0, mf1, m;
		mb1 = p_m[n-1];
		m0 = p_m[n];
		mf1 = p_m[n+1];
		if(n%2){
			mb1 = ~(mb1<<2);
			mf1 = ~(mf1>>3);
		}else{
			mb1 = ~(mb1<<3);
			mf1 = ~(mf1>>2);
		}
		m = mb1 & m0 & mf1;
		if(!m) m = mb1 & m0;
		if(!m) m = m0 & mf1;
		if(!m) m = m0;
		int s, sc[4];
		switch(m){
		case BITS(0,0,0,1,1,1,1,1): goto C5;
		C5:
			sc[0] = p_d[n0-2] < p_d[n0-1] ? -2 : -1;
			sc[1] = p_d[n0] < p_d[n0+1] ? 0 : 1;
			s = p_d[n0+sc[0]] < p_d[n0+sc[1]] ? sc[0] : sc[1];
			s = p_d[n0+s] < p_d[n0+2] ? s : 2;
			break;
		case BITS(0,0,0,1,1,1,1,0): sc[0]= -2; sc[1]= -1; sc[2]=  0; sc[3]=  1; goto C4;
		case BITS(0,0,0,1,1,1,0,1): sc[0]= -2; sc[1]= -1; sc[2]=  0; sc[3]=  2; goto C4;
		case BITS(0,0,0,1,1,0,1,1): sc[0]= -2; sc[1]= -1; sc[2]=  1; sc[3]=  2; goto C4;
		case BITS(0,0,0,1,0,1,1,1): sc[0]= -2; sc[1]=  0; sc[2]=  1; sc[3]=  2; goto C4;
		case BITS(0,0,0,0,1,1,1,1): sc[0]= -1; sc[1]=  0; sc[2]=  1; sc[3]=  2; goto C4;
		C4:
			sc[0] = p_d[n0+sc[0]] < p_d[n0+sc[1]] ? sc[0] : sc[1];
			sc[1] = p_d[n0+sc[2]] < p_d[n0+sc[3]] ? sc[2] : sc[3];
			s = p_d[n0+sc[0]] < p_d[n0+sc[1]] ? sc[0] : sc[1];
			break;
		case BITS(0,0,0,1,1,1,0,0): sc[0]= -2; sc[1]= -1; sc[2]=  0; goto C3;
		case BITS(0,0,0,1,1,0,1,0): sc[0]= -2; sc[1]= -1; sc[2]=  1; goto C3;
		case BITS(0,0,0,1,1,0,0,1): sc[0]= -2; sc[1]= -1; sc[2]=  2; goto C3;
		case BITS(0,0,0,1,0,1,1,0): sc[0]= -2; sc[1]=  0; sc[2]=  1; goto C3;
		case BITS(0,0,0,1,0,1,0,1): sc[0]= -2; sc[1]=  0; sc[2]=  2; goto C3;
		case BITS(0,0,0,1,0,0,1,1): sc[0]= -2; sc[1]=  1; sc[2]=  2; goto C3;
		case BITS(0,0,0,0,1,1,1,0): sc[0]= -1; sc[1]=  0; sc[2]=  1; goto C3;
		case BITS(0,0,0,0,1,1,0,1): sc[0]= -1; sc[1]=  0; sc[2]=  2; goto C3;
		case BITS(0,0,0,0,1,0,1,1): sc[0]= -1; sc[1]=  1; sc[2]=  2; goto C3;
		case BITS(0,0,0,0,0,1,1,1): sc[0]=  0; sc[1]=  1; sc[2]=  2; goto C3;
		C3:
			s = p_d[n0+sc[0]] < p_d[n0+sc[1]] ? sc[0] : sc[1];
			s = p_d[n0+s] < p_d[n0+sc[2]] ? s : sc[2];
			break;
		case BITS(0,0,0,1,1,0,0,0): sc[0]= -2; sc[1]= -1; goto C2;
		case BITS(0,0,0,1,0,1,0,0): sc[0]= -2; sc[1]=  0; goto C2;
		case BITS(0,0,0,1,0,0,1,0): sc[0]= -2; sc[1]=  1; goto C2;
		case BITS(0,0,0,1,0,0,0,1): sc[0]= -2; sc[1]=  2; goto C2;
		case BITS(0,0,0,0,1,1,0,0): sc[0]= -1; sc[1]=  0; goto C2;
		case BITS(0,0,0,0,1,0,1,0): sc[0]= -1; sc[1]=  1; goto C2;
		case BITS(0,0,0,0,1,0,0,1): sc[0]= -1; sc[1]=  2; goto C2;
		case BITS(0,0,0,0,0,1,1,0): sc[0]=  0; sc[1]=  1; goto C2;
		case BITS(0,0,0,0,0,1,0,1): sc[0]=  0; sc[1]=  2; goto C2;
		case BITS(0,0,0,0,0,0,1,1): sc[0]=  1; sc[1]=  2; goto C2;
		C2:
			s = p_d[n0+sc[0]] < p_d[n0+sc[1]] ? sc[0] : sc[1];
			break;
		case BITS(0,0,0,1,0,0,0,0): s= -2; break;
		case BITS(0,0,0,0,1,0,0,0): s= -1; break;
		case BITS(0,0,0,0,0,1,0,0): s=  0; break;
		case BITS(0,0,0,0,0,0,1,0): s=  1; break;
		case BITS(0,0,0,0,0,0,0,1): s=  2; break;
		case BITS(0,0,0,0,0,0,0,0): goto C5;
		default:
			env->ThrowError("Error at %s line %u. m=0x%2.2X, mb1=0x%2.2X, m0=0x%2.2X, mf1=0x%2.2X, n=%u",
				__FILE__, __LINE__, m, mb1, m0, mf1, n);
			break;
		}
		p_ns[n] = n0 + s;
		if(p_ns[n] < 0) p_ns[n] = 0;
		else if(p_ns[n] >= child->GetVideoInfo().num_frames) p_ns[n] = child->GetVideoInfo().num_frames - 1;
	}
	void calc_m(int n, IScriptEnvironment* env){
		if(p_m[n] != 0xFF) return;
		int n0;
		unsigned char bb2, bb1, b0, bf1, bf2;
		
		n0 = n*5/2;
		calc_b(n0-2, env);
		calc_b(n0-1, env);
		calc_b(n0, env);
		calc_b(n0+1, env);
		calc_b(n0+2, env);
		bb2 = p_b[n0-2];
		bb1 = p_b[n0-1];
		b0 = p_b[n0];
		bf1 = p_b[n0+1];
		bf2 = p_b[n0+2];
		
		unsigned int i=0;
		i |= bb2?(1<<5):0;
		i |= bb1?(1<<4):0;
		i |= b0 ?(1<<3):0;
		i |= bf1?(1<<2):0;
		i |= bf2?(1<<1):0;
		i |= (n%2)?1:0;
		
		p_m[n] = phasetable[i];
	}
	void calc_d(int n, IScriptEnvironment* env){
		if(p_d[n] != HUGE_VAL) return;
		calc_d2(n-1, env);
		calc_d2(n, env);
		p_d[n] = p_d2[n]*(1.+blur_rate) - p_d2[n-1]*blur_rate;
	}
	virtual void calc_d2(int n, IScriptEnvironment* env) = 0;
	void calc_b(int n, IScriptEnvironment* env){
		if(p_b[n]!=0xFF) return;
		
//		calc_d(n, env);
//		calc_d(n+1, env);
//		p_b[n] = p_d[n] <= p_d[n+1];
		
//		calc_d(n-1, env);
//		calc_d(n, env);
//		calc_d(n+1, env);
//		p_b[n] = p_d[n] <= (p_d[n-1]+p_d[n+1])/2;
		
		double d[5];
		for(int i=-2; i<=2; i++){
			calc_d(n+i, env);
			d[i+2] = p_d[n+i];
		}
//		calc_d(n-1, env); calc_d(n+1, env);
//		if(p_d[n-1]<p_d[n+1]){
//			for(int i=-4; i<=0; i++){
//				calc_d(n+i, env);
//				d[i+4] = p_d[n+i];
//			}
//		}else{
//			for(int i=0; i<=4; i++){
//				calc_d(n+i, env);
//				d[i] = p_d[n+i];
//			}
//		}
		qsort(d, 5, sizeof(double), calc_b_compare);
		p_b[n] = p_d[n] <= (d[1]+d[3])/2.;
	}
	static int __cdecl calc_b_compare(const void *elem1, const void *elem2){
		return *(double*)elem1 < *(double*)elem2 ? -1 : 1;
	}
#ifdef _TEST
	void write_str(PVideoFrame &c, const char *str, int l=0){
		static const unsigned char font_map[256]={
			0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
			0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
			0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x28,0x00,0x26,0x25,0x00,
			0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0A,0x00,0x00,0x00,0x27,0x00,0x00,
			0x00,0x0B,0x0C,0x0D,0x0E,0x0F,0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,
			0x1A,0x1B,0x1C,0x1D,0x1E,0x1F,0x20,0x21,0x22,0x23,0x24,0x00,0x00,0x00,0x00,0x00,
			0x00,0x0B,0x0C,0x0D,0x0E,0x0F,0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,
			0x1A,0x1B,0x1C,0x1D,0x1E,0x1F,0x20,0x21,0x22,0x23,0x24,0x00,0x00,0x00,0x00,0x00,
			0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
			0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
			0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
			0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
			0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
			0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
			0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
			0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
		};
		if(vi.IsYUY2()){
			unsigned char *dst = c->GetWritePtr();
			unsigned char *dst2;
			const int dst_pitch = c->GetPitch();
			int width = c->GetRowSize() - 3;
			int width2 = (signed)strlen(str)*12*2;
			int height = c->GetHeight() - 1;
			int w, h;
			if(width2 < width) width = width2;
			dst += dst_pitch * l * 12;
			dst2 = dst + dst_pitch;
			height -= l * 12;
			if(12 < height) height = 12;
			for(h=0; h<height; h+=2){
				for(w=0; w<width/4; w++){
					int b;
					char c = str[w/6];
					b = font66[font_map[c]][h/2][w%6];
					if(b){
						dst[w*4] = 0xFF; dst[w*4+1] = 0x80; dst[w*4+2] = 0xFF; dst[w*4+3] = 0x80;
						dst2[w*4] = 0xFF; dst2[w*4+1] = 0x80; dst2[w*4+2] = 0xFF; dst2[w*4+3] = 0x80;
					}else{
						dst[w*4] = 0x00; dst[w*4+1] = 0x80; dst[w*4+2] = 0x00; dst[w*4+3] = 0x80;
						dst2[w*4] = 0x00; dst2[w*4+1] = 0x80; dst2[w*4+2] = 0x00; dst2[w*4+3] = 0x80;
					}
				}
				dst += dst_pitch*2;
				dst2 += dst_pitch*2;
			}
		}else if(vi.IsRGB24()){
			unsigned char *dst = c->GetWritePtr();
			unsigned char *dst2;
			const int dst_pitch = c->GetPitch();
			int width = c->GetRowSize() - 3;
			int width2 = (signed)strlen(str)*12*3;
			int height = c->GetHeight() - 1;
			int w, h;
			if(width2 < width) width = width2;
			dst += height * dst_pitch;
			dst -= dst_pitch * l * 12;
			dst2 = dst - dst_pitch;
			height -= l * 12;
			if(12 < height) height = 12;
			for(h=0; h<height; h+=2){
				for(w=0; w<width/6; w++){
					int b;
					char c = str[w/6];
					b = font66[font_map[c]][h/2][w%6];
					if(b){
						dst[w*6] = 0xFF; dst[w*6+1] = 0xFF; dst[w*6+2] = 0xFF; dst[w*6+3] = 0xFF; dst[w*6+4] = 0xFF; dst[w*6+5] = 0xFF;
						dst2[w*6] = 0xFF; dst2[w*6+1] = 0xFF; dst2[w*6+2] = 0xFF; dst2[w*6+3] = 0xFF; dst2[w*6+4] = 0xFF; dst2[w*6+5] = 0xFF;
					}else{
						dst[w*6] = 0x00; dst[w*6+1] = 0x00; dst[w*6+2] = 0x00; dst[w*6+3] = 0x00; dst[w*6+4] = 0x00; dst[w*6+5] = 0x00;
						dst2[w*6] = 0x00; dst2[w*6+1] = 0x00; dst2[w*6+2] = 0x00; dst2[w*6+3] = 0x00; dst2[w*6+4] = 0x00; dst2[w*6+5] = 0x00;
					}
				}
				dst -= dst_pitch*2;
				dst2 -= dst_pitch*2;
			}
		}else if(vi.IsRGB32()){
			unsigned char *dst = c->GetWritePtr();
			unsigned char *dst2;
			const int dst_pitch = c->GetPitch();
			int width = c->GetRowSize() - 3;
			int width2 = (signed)strlen(str)*12*4;
			int height = c->GetHeight() - 1;
			int w, h;
			if(width2 < width) width = width2;
			dst += height * dst_pitch;
			dst -= dst_pitch * l * 12;
			dst2 = dst - dst_pitch;
			height -= l * 12;
			if(12 < height) height = 12;
			for(h=0; h<height; h+=2){
				for(w=0; w<width/8; w++){
					int b;
					char c = str[w/6];
					b = font66[font_map[c]][h/2][w%6];
					if(b){
						dst[w*8] = 0xFF; dst[w*8+1] = 0xFF; dst[w*8+2] = 0xFF; dst[w*8+4] = 0xFF; dst[w*8+5] = 0xFF; dst[w*8+6] = 0xFF;
						dst2[w*8] = 0xFF; dst2[w*8+1] = 0xFF; dst2[w*8+2] = 0xFF; dst2[w*8+4] = 0xFF; dst2[w*8+5] = 0xFF; dst2[w*8+6] = 0xFF;
					}else{
						dst[w*8] = 0x00; dst[w*8+1] = 0x00; dst[w*8+2] = 0x00; dst[w*8+4] = 0x00; dst[w*8+5] = 0x00; dst[w*8+6] = 0x00;
						dst2[w*8] = 0x00; dst2[w*8+1] = 0x00; dst2[w*8+2] = 0x00; dst2[w*8+4] = 0x00; dst2[w*8+5] = 0x00; dst2[w*8+6] = 0x00;
					}
				}
				dst -= dst_pitch*2;
				dst2 -= dst_pitch*2;
			}
		}else if(vi.IsYV12()){
			unsigned char *dst = c->GetWritePtr();
			unsigned char *dst2;
			unsigned char *dstu = c->GetWritePtr(PLANAR_U);
			unsigned char *dstv = c->GetWritePtr(PLANAR_V);
			const int dst_pitch = c->GetPitch();
			const int dstuv_pitch = c->GetPitch(PLANAR_U);
			int width = c->GetRowSize() - 1;
			int width2 = (signed)strlen(str)*12;
			int height = c->GetHeight() - 1;
			int w, h;
			if(width2 < width) width = width2;
			dst += dst_pitch * l * 12;
			dstu += dstuv_pitch * l *6;
			dstv += dstuv_pitch * l *6;
			dst2 = dst + dst_pitch;
			height -= l * 12;
			if(12 < height) height = 12;
			for(h=0; h<height; h+=2){
				for(w=0; w<width/2; w++){
					int b;
					char c = str[w/6];
					b = font66[font_map[c]][h/2][w%6];
					if(b){
						dst[w*2] = 0xFF; dst[w*2+1] = 0xFF;
						dst2[w*2] = 0xFF; dst2[w*2+1] = 0xFF;
						dstu[w] = 0x80; dstv[w] = 0x80;
					}else{
						dst[w*2] = 0x00; dst[w*2+1] = 0x00;
						dst2[w*2] = 0x00; dst2[w*2+1] = 0x00;
						dstu[w] = 0x80; dstv[w] = 0x80;
					}
				}
				dst += dst_pitch*2;
				dst2 += dst_pitch*2;
				dstu += dstuv_pitch;
				dstv += dstuv_pitch;
			}
		}
	}
#endif
};

template <int x_step, int y_step, int xpitch>
class a60224 : public a60224_common{
public:
	a60224(
		PClip _child, double _blur_rate,
		int _crop_x1, int _crop_x2, int _crop_y1, int _crop_y2
	) : a60224_common(
		_child, _blur_rate,
		_crop_x1, _crop_x2, _crop_y1, _crop_y2
	){}
protected:
	void calc_d2(int n, IScriptEnvironment* env){
		if(p_d2[n] != HUGE_VAL) return;
		PVideoFrame c;
		c = child->GetFrame(n, env);
		_asm{ emms }
		set_cache(c);
		const unsigned char* cp = c->GetReadPtr();
		const int c_pitch = c->GetPitch();
		const int c_width = c->GetRowSize();
		const int c_height = c->GetHeight();
		int w, h;
		
		//double d_all = 0.;
		__int64 d_all = 0;
		unsigned long d_line;
		long d1, d2, d3, D, dl, ds;
		cp += c_pitch;
		const unsigned char *cpu = cp - c_pitch;
		const unsigned char *cpd = cp + c_pitch;
		for(h=1+crop_y1; h<c_height-1-crop_y2; h+=y_step){
			d_line = 0;
			for(w=crop_x1*xpitch; w<c_width-crop_x2*xpitch; w+=x_step){
				long yu, y0, yd;
				yu = cpu[w]; y0 = cp[w]; yd = cpd[w]; w+=x_step;
				d1 = abs(y0-yu);
				d2 = abs(yd-y0);
				d3 = abs(yu-yd);
				if(d1 < d2){
					ds = d1; dl = d2;
				}else{
					dl = d1; ds = d2;
				}
				if(d3 < ds){
					D = dl-d3;
				}else{
					D = dl-ds; //xcXN[΍
				}
				d_line += D*D; //66051sNZ܂Ł255*255*66052>2^32-1
			}
			d_all += d_line;
			cp += c_pitch*y_step;
			cpu += c_pitch*y_step;
			cpd += c_pitch*y_step;
		}
		p_d2[n] = sqrt((double)d_all);
	}
};

AVSValue __cdecl Create(AVSValue args, void* user_data, IScriptEnvironment* env){
	PClip child = args[0].AsClip();
	bool fast = args[1].AsBool(false);
	double blur_rate = args[2].AsFloat(0.);
	int x1 = args[3].AsInt(0);
	int x2 = args[4].AsInt(0);
	int y1 = args[5].AsInt(0);
	int y2 = args[6].AsInt(0);
	const VideoInfo& vi = child->GetVideoInfo();
	if(fast){
		if(vi.IsYUY2()){
			return new a60224<4,2,2>(child, blur_rate, x1, x2, y1, y2);
		}
		if(vi.IsYV12()){
			return new a60224<2,2,1>(child, blur_rate, x1, x2, y1, y2);
		}
		if(vi.IsRGB24()){
			return new a60224<2,2,3>(child, blur_rate, x1, x2, y1, y2);
		}
		if(vi.IsRGB32()){
			return new a60224<2,2,4>(child, blur_rate, x1, x2, y1, y2);
		}
	}else{
		if(vi.IsYUY2()){
			return new a60224<2,1,2>(child, blur_rate, x1, x2, y1, y2);
		}
		if(vi.IsYV12()){
			return new a60224<1,1,1>(child, blur_rate, x1, x2, y1, y2);
		}
		if(vi.IsRGB24()){
			return new a60224<1,1,3>(child, blur_rate, x1, x2, y1, y2);
		}
		if(vi.IsRGB32()){
			return new a60224<1,1,4>(child, blur_rate, x1, x2, y1, y2);
		}
	}
	env->ThrowError("cc: Unsupported color format.");
	__assume(0);
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env) {
	env->AddFunction("a60224", "c[fast]b[blur_rate]f[x1]i[x2]i[y1]i[y2]i", Create, 0);
	return "`a60224' by chiyo-clone.";
}
